import torch
import torch.nn as nn
import numpy as np

from aenet_gpr.src.prior import ConstantPrior
from aenet_gpr.src.pytorch_kernel import FPKernel, FPKernelNoforces


class GaussianProcess(nn.Module):
    '''
    Gaussian Process Regression
    Parameters:

    prior: Defaults to ConstantPrior with zero as constant

    kernel: Defaults to the Squared Exponential kernel with derivatives
    '''

    def __init__(self, hp=None, prior=None, prior_update=True, kerneltype='sqexp',
                 scale=0.4, weight=1.0, noise=1e-6, noisefactor=0.5,
                 use_forces=True, images=None, function=None, derivative=None,
                 sparse=None, sparse_derivative=None, autograd=False,
                 data_type='float64', device='cpu', soap_param=None, descriptor='cartesian coordinates',
                 atoms_mask=None):
        super().__init__()

        if data_type == 'float32':
            self.data_type = 'float32'
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.torch_data_type = torch.float64

        self.device = device
        self.soap_param = soap_param
        self.descriptor = descriptor
        self.kerneltype = kerneltype

        if autograd:
            self.scale = nn.Parameter(torch.tensor(scale, dtype=self.torch_data_type), requires_grad=True).to(self.device)
            self.weight = nn.Parameter(torch.tensor(weight, dtype=self.torch_data_type), requires_grad=True).to(self.device)
        else:
            self.scale = torch.tensor(scale, dtype=self.torch_data_type, device=self.device)
            self.weight = torch.tensor(weight, dtype=self.torch_data_type, device=self.device)

        self.noise = torch.tensor(noise, dtype=self.torch_data_type, device=self.device)
        self.noisefactor = torch.tensor(noisefactor, dtype=self.torch_data_type, device=self.device)

        self.use_forces = use_forces
        self.images = images
        self.Ntrain = len(self.images)
        self.species = self.images[0].get_chemical_symbols()
        self.pbc = np.all(self.images[0].get_pbc())

        self.Natom = len(self.species)
        self.atoms_mask = atoms_mask

        if self.atoms_mask is not None:
            self.Nmask = self.atoms_mask.shape[0]
        else:
            self.Nmask = 3 * self.Natom

        self.Y = function  # Y = [Ntrain]
        self.dY = derivative  # dY = [Ntrain, Natom, 3]
        self.model_vector = torch.empty((self.Ntrain * (1 + self.Nmask),), dtype=self.torch_data_type, device=self.device)

        if prior is None:
            self.prior = ConstantPrior(0.0, dtype=self.torch_data_type, device=self.device, atoms_mask=self.atoms_mask)
        else:
            self.prior = torch.tensor(prior, dtype=self.torch_data_type, device=self.device)
        self.prior_update = prior_update

        self.sparse = sparse

        if sparse is not None:
            self.sX = sparse  # sX = [Nsparse, Nscenter, Nfeature]
            self.sparse = True

            if sparse_derivative is not None:
                self.sdX = sparse_derivative  # sdX = [Nsparse, Nscenter, Natom, 3, Nfeature]
            else:
                self.sdX = None

        else:
            self.sX = None
            self.sparse = False

        if self.use_forces:
            self.kernel = FPKernel(species=self.species,
                                   pbc=self.pbc,
                                   Natom=self.Natom,
                                   kerneltype=self.kerneltype,
                                   data_type=self.data_type,
                                   soap_param=self.soap_param,
                                   descriptor=self.descriptor,
                                   device=self.device,
                                   atoms_mask=self.atoms_mask)
        else:
            self.kernel = FPKernelNoforces(species=self.species,
                                           pbc=self.pbc,
                                           Natom=self.Natom,
                                           kerneltype=self.kerneltype,
                                           data_type=self.data_type,
                                           soap_param=self.soap_param,
                                           descriptor=self.descriptor,
                                           device=self.device,
                                           atoms_mask=self.atoms_mask)

        hyper_params = dict(kerneltype=self.kerneltype,
                            scale=self.scale,
                            weight=self.weight,
                            noise=self.noise,
                            noisefactor=self.noisefactor,
                            prior=self.prior.constant)

        self.hyper_params = hyper_params
        self.kernel.set_params(self.hyper_params)

        if self.Y is not None:
            if self.dY is not None:
                # [Ntrain] -> [Ntrain, 1]
                # Y_reshaped = self.Y.flatten().unsqueeze(1)
                Y_reshaped = self.Y.contiguous().view(-1, 1)

                # [Ntrain, Natom, 3] -> [Ntrain * 3 * Natom, 1]
                # dY_reshaped = self.dY.flatten().unsqueeze(1)
                if self.atoms_mask is not None:
                    dY_reshaped = self.dY.contiguous().view(self.dY.shape[0], -1)[:, self.atoms_mask]
                    dY_reshaped = dY_reshaped.view(-1, 1)
                else:
                    dY_reshaped = self.dY.contiguous().view(-1, 1)

                # [Ntrain * (1 + 3 * Natom), 1]
                # [[e1, e2, ..., eN, f11x, f11y, f11z, f12x, f12y, ..., fNzNz]],
                self.YdY = torch.cat((Y_reshaped, dY_reshaped), dim=0)

                del Y_reshaped, dY_reshaped

            else:
                self.YdY = self.Y.flatten().unsqueeze(1)  # no dY [e1, e2, ..., eN]
        else:
            self.YdY = None

    def train_model(self):
        if self.sparse:
            if self.use_forces:  # self.kernel = FPKernel
                # covariance matrix between the training points X
                K_XX = self.kernel.kernel_matrix_iterative(images=self.images)

                # reg = [Ntrain * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)]
                a = torch.tensor(self.Ntrain * [self.hyper_params['noise'] * self.hyper_params['noisefactor']], dtype=self.torch_data_type, device=self.device).reshape(
                    self.Ntrain, 1)
                b = torch.tensor(self.Ntrain * self.Nmask * [self.hyper_params['noise']],
                                 dtype=self.torch_data_type, device=self.device).reshape(self.Ntrain, -1)
                reg = torch.diag(torch.cat((a, b), 1).flatten() ** 2)
                self.inv_reg = torch.linalg.inv(reg)

                K_XX.add_(reg)
                try:
                    self.K_XX_L = torch.linalg.cholesky(K_XX, upper=False)
                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(K_XX))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(K_XX.shape[0],
                                                                            dtype=self.torch_data_type)
                        K_XX.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_XX after adjusting
                        self.K_XX_L = torch.linalg.cholesky(K_XX, upper=False)

                # covariance matrix between the inducing points S
                K_ss = self.kernel.kernel_matrix_iterative(images=self.images)

                try:
                    # Step 1: Cholesky decomposition for K_ss
                    self.K_ss_L = torch.linalg.cholesky(K_ss, upper=False)
                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(K_ss))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(K_ss.shape[0],
                                                                            dtype=self.torch_data_type)
                        K_ss.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_ss after adjusting
                        self.K_ss_L = torch.linalg.cholesky(K_ss, upper=False)

                # covariance between inducing points S and training points X
                self.K_sX = self.kernel.kernel_vector_iterative(x=self.sX, dx=self.sdX, X=self.X, dX=self.dX)

                Qs = K_ss + torch.einsum('pi,ij,jq->pq', self.K_sX, self.inv_reg, self.K_sX.T)

                CKK_XX_L = torch.cholesky_solve(self.K_sX.T.clone(), self.K_XX_L, upper=False)
                Q_ss = torch.einsum('ij,jk->ik', self.K_sX, CKK_XX_L)
                self.CQ_ss = torch.cholesky_solve(Q_ss, self.K_ss_L, upper=False)

            else:  # self.kernel = FPKernelNoforces
                # covariance matrix between the training points X
                K_XX = self.kernel.kernel_without_deriv(X1=self.X, X2=self.X)

                a = torch.tensor(self.Ntrain * [self.hyper_params['noise']], dtype=self.torch_data_type, device=self.device)
                reg = torch.diag(a ** 2)
                self.inv_reg = torch.linalg.inv(reg)

                K_XX.add_(reg)
                try:
                    self.K_XX_L = torch.linalg.cholesky(K_XX, upper=False)
                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(K_XX))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(K_XX.shape[0],
                                                                            dtype=self.torch_data_type)
                        K_XX.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_XX after adjusting
                        self.K_XX_L = torch.linalg.cholesky(K_XX, upper=False)

                # covariance matrix between the inducing points S
                K_ss = self.kernel.kernel_without_deriv(X1=self.sX, X2=self.sX)

                try:
                    # Step 1: Cholesky decomposition for K_ss
                    self.K_ss_L = torch.linalg.cholesky(K_ss, upper=False)
                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(K_ss))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(K_ss.shape[0],
                                                                            dtype=self.torch_data_type)
                        K_ss.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_ss after adjusting
                        self.K_ss_L = torch.linalg.cholesky(K_ss, upper=False)

                # covariance between inducing points S and training points X
                self.K_sX = self.kernel.kernel_without_deriv(X1=self.sX, X2=self.X)

                # KK = [Nsparse, Nsparse]
                Qs = K_ss + torch.einsum('pi,ij,jq->pq', self.K_sX, self.inv_reg, self.K_sX.T)

                CKK_XX_L = torch.cholesky_solve(self.K_sX.T.clone(), self.K_XX_L, upper=False)
                Q_ss = torch.einsum('ij,jk->ik', self.K_sX, CKK_XX_L)
                self.CQ_ss = torch.cholesky_solve(Q_ss, self.K_ss_L, upper=False)

            self.model_vector = self.calculate_model_vector_sparse(matrix=Qs)

        else:
            if self.use_forces:  # self.kernel = FPKernel
                # covariance matrix between the training points X
                self.K_XX_L = self.kernel.kernel_matrix_iterative(images=self.images)

                a = torch.full((self.Ntrain, 1), self.hyper_params['noise'] * self.hyper_params['noisefactor'],
                               dtype=self.torch_data_type, device=self.device)
                noise_val = self.hyper_params['noise']
                b = noise_val.expand(self.Ntrain, self.Nmask)

                # reg = torch.diag(torch.cat((a, b), 1).flatten() ** 2)
                diagonal_values = torch.cat((a, b), 1).flatten() ** 2

                self.K_XX_L.diagonal().add_(diagonal_values)

                try:
                    self.K_XX_L = torch.linalg.cholesky(self.K_XX_L, upper=False)

                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(self.K_XX_L))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(self.K_XX_L.shape[0],
                                                                            dtype=self.torch_data_type,
                                                                            device=self.device)
                        self.K_XX_L.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_XX after adjusting
                        self.K_XX_L = torch.linalg.cholesky(self.K_XX_L, upper=False)

            else:  # self.kernel = FPKernelNoforces
                # KK = [Ntrain, Ntrain]
                __K_XX = self.kernel.kernel_matrix(X=self.X)

                a = torch.tensor(self.Ntrain * [self.hyper_params['noise']], dtype=self.torch_data_type, device=self.device)
                reg = torch.diag(a ** 2)

                __K_XX.add_(reg)
                try:
                    self.K_XX_L = torch.linalg.cholesky(__K_XX, upper=False)
                except torch.linalg.LinAlgError:
                    with torch.no_grad():
                        # Diagonal sum (trace)
                        diag_sum = torch.sum(torch.diag(__K_XX))

                        # epsilon value
                        eps = torch.finfo(self.torch_data_type).eps

                        # scaling factor
                        scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                        # adjust K_XX
                        adjustment = diag_sum * scaling_factor * torch.ones(__K_XX.shape[0],
                                                                            dtype=self.torch_data_type,
                                                                            device=self.device)
                        __K_XX.diagonal().add_(adjustment)

                        # Step 1: Cholesky decomposition for K_XX after adjusting
                        self.K_XX_L = torch.linalg.cholesky(__K_XX, upper=False)

            if self.prior_update:
                self.prior.update(self.images, self.YdY, self.K_XX_L, self.use_forces)
                self.hyper_params.update(dict(prior=self.prior.constant))
                self.kernel.set_params(self.hyper_params)

            _prior_array = self.prior.potential_batch(self.images, self.use_forces)
            self.model_vector = torch.cholesky_solve(self.YdY.contiguous().view(-1, 1) - _prior_array.view(-1, 1),
                                                     self.K_XX_L, upper=False)

        return

    def calculate_model_vector(self, matrix):
        """
        self.YdY.shape  # [Ntrain, 1 + 3 * Natom]
        model_vector.shape  # [Ntrain * (1 + 3 * Natom)]
        model_vector.unsqueeze(1).shape  # [Ntrain * (1 + 3 * Natom), 1]
        self.K_XX_L.shape  # [Ntrain * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)]
        """

        # Factorize K-matrix (Cholesky decomposition) using torch.linalg.cholesky:
        self.K_XX_L = torch.linalg.cholesky(matrix, upper=False)  # Lower triangular by default

        # Flatten Y and compute the model vector
        model_vector = self.YdY.flatten()  # - self.prior_array

        # Solve the system L * L^T * v = model_vector (Cholesky solve)
        model_vector = torch.cholesky_solve(model_vector.unsqueeze(1), self.K_XX_L, upper=False)

        return model_vector

    def calculate_model_vector_sparse(self, matrix):
        # [Nsparse * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)] * 
        # [Ntrain * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)] *
        # [Ntrain * (1 + 3 * Natom), 1]
        # -> [Nsparse * (1 + 3 * Natom), 1]
        reduced_target = torch.einsum('pi,ij,jq->pq', self.K_sX, self.inv_reg, self.YdY.reshape(-1, 1))

        try:
            matrix_L = torch.linalg.cholesky(matrix, upper=False)
        except torch.linalg.LinAlgError:
            with torch.no_grad():
                # Diagonal sum (trace)
                diag_sum = torch.sum(torch.diag(matrix))

                # epsilon value
                eps = torch.finfo(self.torch_data_type).eps

                # scaling factor
                scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                # adjust K_XX
                adjustment = diag_sum * scaling_factor * torch.ones(matrix.shape[0],
                                                                    dtype=self.torch_data_type)
                matrix.diagonal().add_(adjustment)

                # Step 1: Cholesky decomposition for K_XX after adjusting
                matrix_L = torch.linalg.cholesky(matrix, upper=False)

        # Solve the system L * L^T * v = model_vector (Cholesky solve)
        model_vector = torch.cholesky_solve(reduced_target, matrix_L, upper=False)

        return model_vector

    def forward(self, eval_images, get_variance=False):

        Ntest = len(eval_images)
        if not get_variance:
            if self.use_forces:
                # E_hat = pred[0:x.shape[0]]
                # F_hat = pred[x.shape[0]:].reshape(x.shape[0], self.Natom, -1)
                E_hat = torch.empty((Ntest,), dtype=self.torch_data_type, device=self.device)
                F_hat = torch.zeros((Ntest, self.Natom * 3), dtype=self.torch_data_type, device=self.device)

                for i, eval_image in enumerate(eval_images):
                    pred, kernel = self.eval_data_per_data(eval_image=eval_image)

                    E_hat[i] = pred[0]
                    if self.atoms_mask is not None:
                        F_hat[i, self.atoms_mask] = pred[1:].view(-1)
                    else:
                        F_hat[i, :] = pred[1:].view(-1)

                return E_hat, F_hat.view((Ntest, self.Natom, 3)), None

            else:
                pass
                # E_hat = pred
                #
                # return E_hat, None, None

        else:
            if self.use_forces:
                E_hat = torch.empty((Ntest,), dtype=self.torch_data_type, device=self.device)
                F_hat = torch.zeros((Ntest, self.Natom * 3), dtype=self.torch_data_type, device=self.device)
                uncertainty = torch.empty((Ntest,), dtype=self.torch_data_type, device=self.device)

                for i, eval_image in enumerate(eval_images):
                    pred, kernel = self.eval_data_per_data(eval_image=eval_image)
                    var = self.eval_variance_per_data(get_variance=get_variance, eval_image=eval_image, k=kernel)

                    E_hat[i] = pred[0]
                    if self.atoms_mask is not None:
                        F_hat[i, self.atoms_mask] = pred[1:].view(-1)
                    else:
                        F_hat[i, :] = pred[1:].view(-1)
                    uncertainty[i] = torch.sqrt(torch.diagonal(var)[0]) / self.weight

                return E_hat, F_hat((Ntest, self.Natom, 3)), uncertainty

            else:
                pass
                # E_hat = pred
                # uncertainty_squared = torch.diagonal(var)
                # uncertainty = torch.sqrt(uncertainty_squared)
                #
                # return E_hat, None, uncertainty

    def eval_data_per_data(self, eval_image):

        # kernel between test point x and inducing points S
        if self.sparse:
            if self.use_forces:
                kernel = self.kernel.kernel_vector_per_data(eval_image=eval_image, train_images=self.images)
            # else:
            #     kernel = self.kernel.kernel_vector(x=x_i, X=self.sX)

            # pred = torch.einsum('hi,i->h', kernel, self.model_vector.flatten())
            pred = torch.matmul(kernel, self.model_vector.view(-1))

        # kernel between test point x and training points X
        else:
            if self.use_forces:
                kernel = self.kernel.kernel_vector_per_data(eval_image=eval_image, train_images=self.images)
            # else:
            #     kernel = self.kernel.kernel_vector(x=x_i, X=self.X)

            # pred = torch.einsum('hi,i->h', kernel, self.model_vector.flatten())
            pred = torch.matmul(kernel, self.model_vector.view(-1)) + self.prior.potential_per_data(eval_image, self.use_forces)

        return pred, kernel

    def eval_variance_per_data(self, get_variance, eval_image, k):
        """
        variance.shape  # [Ntest * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        self.K_XX_L.shape  # [Ntrain * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)]
        k.T.clone().shape  # [Ntrain * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        self.Ck.shape  # [Ntrain * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        covariance.shape  # [Ntest * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        """
        # var = None
        if get_variance:
            # Compute variance of test points x
            # if self.use_forces:
            #     variance = self.kernel.kernel_matrix_iterative(X=x, dX=dx)
            # else:
            #     variance = self.kernel.kernel_matrix(X=x)

            # Perform Cholesky decomposition and solve the system
            if self.sparse:
                # Step 2: Cholesky solve
                CK_ss_L = torch.cholesky_solve(k.T.clone(), self.K_ss_L, upper=False)

                covariance = torch.einsum('pi,ij,jq->pq', k,
                                          torch.eye(self.CQ_ss.shape[0], dtype=self.torch_data_type) - self.CQ_ss,
                                          CK_ss_L)

            else:
                # CK_XX_L = torch.cholesky_solve(k.T.clone(), self.K_XX_L, upper=False)
                # covariance = torch.einsum('ij,jk->ik', k, CK_XX_L)
                covariance = torch.matmul(k, torch.cholesky_solve(k.T.clone(), self.K_XX_L, upper=False))

            # Adjust variance by subtracting covariance
            if self.use_forces:
                return self.kernel.kernel_matrix_per_data(image=eval_image) - covariance
            # else:
            #     return self.kernel.kernel_matrix(X=x_i) - covariance

        else:
            return None

    def save_data(self, file="calc_dict.pt"):
        """
        self.data_type
        self.torch_data_type

        self.device = device
        self.noise
        self.noisefactor
        self.scale
        self.weight
        self.use_forces
        self.sparse

        self.images
        self.Y
        self.dY
        self.YdY

        self.Ntrain
        self.Natom

        self.K_XX_L
        self.model_vector
        """

        state = {
            'kerneltype': self.kerneltype,
            'noise': self.noise,
            'noisefactor': self.noisefactor,
            'scale': self.scale,
            'weight': self.weight,
            'use_forces': self.use_forces,
            'sparse': self.sparse,
            'Y': self.Y,
            'dY': self.dY,
            'YdY': self.YdY,
            'K_XX_L': self.K_XX_L,
            'model_vector': self.model_vector,
        }
        torch.save(state, file)

    def load_data(self, file="calc_dict.pt"):
        state = torch.load(file)

        self.kerneltype = state.get('kerneltype')
        self.noise = state.get('noise')
        self.noisefactor = state.get('noisefactor')
        self.scale = state.get('scale')
        self.weight = state.get('weight')

        self.use_forces = state.get('use_forces')
        self.sparse = state.get('sparse')

        self.Y = state.get('Y')
        self.dY = state.get('dY')
        self.YdY = state.get('YdY')

        self.K_XX_L = state.get('K_XX_L')
        self.model_vector = state.get('model_vector')

        hyper_params = dict(kerneltype=self.kerneltype,
                            scale=self.scale,
                            weight=self.weight,
                            noise=self.noise,
                            noisefactor=self.noisefactor)

        self.hyper_params = hyper_params
        self.kernel.set_params(self.hyper_params)



import torch
import numpy as np
import os

# import chemcoord as cc
import ase.io

from aenet_gpr.src.pytorch_kerneltypes import SquaredExp
from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch


class BaseKernelType:
    '''
    Base class for all kernel types with common properties,
    attributes and methods.
    '''

    def __init__(self, params=None):
        # Currently, all the kernel types take weight and scale as parameters

        if params is None:
            self.params = {}

    @property
    def scale(self):
        return self.params['scale']

    @property
    def weight(self):
        return self.params['weight']

    def update(self, params):
        '''
        Update the kernel function hyperparameters.
        '''

        self.params.update(params)


class FPKernelInternal(BaseKernelType):

    def __init__(self,
                 species, pbc, Natom,
                 kerneltype='sqexp',
                 params=None,
                 data_type='float64',
                 soap_param=None,
                 descriptor='internal',
                 device='cpu',
                 atoms_mask=None,
                 c_table=None):
        super().__init__()
        '''
        params: dict
            Hyperparameters for the kernel type
        '''
        kerneltypes = {'sqexp': SquaredExp}
        self.device = device
        self.atoms_mask = atoms_mask

        if data_type == 'float32':
            self.data_type = 'float32'
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.torch_data_type = torch.float64

        self.species = species
        self.pbc = pbc
        self.Natom = Natom
        self.c_table = c_table

        if self.atoms_mask is not None:
            self.Nmask = self.atoms_mask.shape[0]
            self.free_atoms = self.atoms_mask // 3
            self.free_atoms = torch.unique(self.free_atoms)
        else:
            self.Nmask = 3 * self.Natom

        if params is None:
            params = {}

        kerneltype = kerneltypes.get(kerneltype)
        self.kerneltype = kerneltype(**params)

        self.soap_param = soap_param
        self.descriptor = descriptor
        self.soap = None

    def pairwise_distances(self, fingerprints, sin_transform=False):
        # fingerprints: (Ndata, Ncenter, Nfeature)
        Ndata = fingerprints.shape[0]

        distances = []
        for i in range(Ndata):
            for j in range(i + 1, Ndata):
                diff = fingerprints[i].flatten() - fingerprints[j].flatten()
                if sin_transform:
                    diff = torch.sin(diff / 2)
                distance = torch.linalg.norm(diff)
                distances.append(distance.item())

        distances = torch.tensor(distances)
        mean_distance = torch.mean(distances)
        std_distance = torch.std(distances)
        max_distance = torch.max(distances)
        min_distance = torch.min(distances)

        return mean_distance, std_distance, max_distance, min_distance

    def generate_internal_descriptor(self, images):

        Ndata = len(images)
        Ncenter = 1
        Natom = len(images[0])
        Nfeature = Natom

        fp = {'bond': torch.empty((Ndata, Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device),
              'angle': torch.empty((Ndata, Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device),
              'dihedral': torch.empty((Ndata, Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device), }

        dfp_dr = {'bond': torch.empty((Ndata, Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device),
                  'angle': torch.empty((Ndata, Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device),
                  'dihedral': torch.empty((Ndata, Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device), }

        # Set the c_table
        if self.c_table is None:
            ase.io.write("./tmp.xyz", images[0], plain=True)
            cc_atom = cc.Cartesian.read_xyz('./tmp.xyz', start_index=1)
            os.remove("./tmp.xyz")

            z_matrix = cc_atom.get_zmat()
            self.c_table = z_matrix.loc[:, ['b', 'a', 'd']]

            print("c_table has been constructed:")
            print(self.c_table)

        # Set z-matrix and z-matrix-gradient
        for i, image in enumerate(images):
            ase.io.write("./tmp_{0}.xyz".format(i), image, plain=True)
            cc_atom = cc.Cartesian.read_xyz('./tmp_{0}.xyz'.format(i), start_index=1)
            os.remove("./tmp_{0}.xyz".format(i))

            # z-matrix
            zmat = cc_atom.get_zmat(self.c_table)

            fp.get('bond')[i, 0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'bond']),
                                                      dtype=self.torch_data_type).to(self.device)
            fp.get('angle')[i, 0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'angle']),
                                                       dtype=self.torch_data_type).to(self.device)
            fp.get('dihedral')[i, 0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'dihedral']),
                                                          dtype=self.torch_data_type).to(self.device)

            # z-matrix-gradient
            zmat_grad = cc_atom.loc[self.c_table.index].get_grad_zmat(self.c_table).args[0]
            zmat_grad = torch.as_tensor(zmat_grad, dtype=self.torch_data_type).to(self.device)

            dfp_dr.get('bond')[i, 0, :, :, :] = zmat_grad[0].permute(1, 2, 0)  # 0, 2, 1
            dfp_dr.get('angle')[i, 0, :, :, :] = zmat_grad[1].permute(1, 2, 0)
            dfp_dr.get('dihedral')[i, 0, :, :, :] = zmat_grad[2].permute(1, 2, 0)

        return fp, dfp_dr

    def generate_internal_descriptor_per_data(self, image):

        Ncenter = 1
        Natom = len(image)
        Nfeature = Natom

        fp = {'bond': torch.empty((Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device),
              'angle': torch.empty((Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device),
              'dihedral': torch.empty((Ncenter, Nfeature), dtype=self.torch_data_type, device=self.device), }

        dfp_dr = {'bond': torch.empty((Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device),
                  'angle': torch.empty((Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device),
                  'dihedral': torch.empty((Ncenter, Natom, 3, Nfeature), dtype=self.torch_data_type, device=self.device), }

        # Set the c_table
        if self.c_table is None:
            ase.io.write("./tmp.xyz", image, plain=True)
            cc_atom = cc.Cartesian.read_xyz('./tmp.xyz', start_index=1)

            z_matrix = cc_atom.get_zmat()
            self.c_table = z_matrix.loc[:, ['b', 'a', 'd']]

        cc_atom = cc.Cartesian.read_xyz('./tmp.xyz', start_index=1)
        os.remove("./tmp.xyz")

        # z-matrix
        zmat = cc_atom.get_zmat(self.c_table)

        fp.get('bond')[0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'bond']),
                                               dtype=self.torch_data_type).to(self.device)
        fp.get('angle')[0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'angle']),
                                                dtype=self.torch_data_type).to(self.device)
        fp.get('dihedral')[0, :] = torch.as_tensor(np.array(zmat.safe_loc[:, 'dihedral']),
                                                   dtype=self.torch_data_type).to(self.device)

        # z-matrix-gradient
        zmat_grad = cc_atom.loc[self.c_table.index].get_grad_zmat(self.c_table).args[0]
        zmat_grad = torch.as_tensor(zmat_grad, dtype=self.torch_data_type).to(self.device)

        dfp_dr.get('bond')[0, :, :, :] = zmat_grad[0].permute(1, 2, 0)  # 0, 2, 1
        dfp_dr.get('angle')[0, :, :, :] = zmat_grad[1].permute(1, 2, 0)
        dfp_dr.get('dihedral')[0, :, :, :] = zmat_grad[2].permute(1, 2, 0)

        return fp, dfp_dr

    def kernel_with_deriv(self, X1=None, dX1=None, X2=None, dX2=None):
        '''
        Return a full kernel matrix between two structural fingerprints, 'X1' and 'X2'.
        X1, X2 = [Ndata, Ncenter, Nfeature]
        dX1, dX2 = [Ndata, Ncenter, Natom, 3, Nfeature]

        where Ndata = Nbatch, Ncenter=1, Nfeature=Natom
        '''

        X1_b = X1.get('bond')
        X1_a = X1.get('angle') * torch.pi / 180  # unit: degree -> radian
        X1_d = X1.get('dihedral') * torch.pi / 180  # unit: degree -> radian
        dX1_b = dX1.get('bond')
        dX1_a = dX1.get('angle')  # unit: radian
        dX1_d = dX1.get('dihedral')  # unit: radian

        X2_b = X2.get('bond')
        X2_a = X2.get('angle') * torch.pi / 180  # unit: degree -> radian
        X2_d = X2.get('dihedral') * torch.pi / 180  # unit: degree -> radian
        dX2_b = dX2.get('bond')
        dX2_a = dX2.get('angle')  # unit: radian
        dX2_d = dX2.get('dihedral')  # unit: radian

        scale_b = self.scale.get('bond')
        scale_a = self.scale.get('angle')
        scale_d = self.scale.get('dihedral')

        # pairwise_distances example
        # mean_bond, std_bond, max_bond, min_bond = self.pairwise_distances(X1_b)
        # mean_angle_sin, std_angle_sin, max_angle_sin, min_angle_sin = self.pairwise_distances(X1_a, sin_transform=True)
        # mean_dihedral_sin, std_dihedral_sin, max_dihedral_sin, min_dihedral_sin = self.pairwise_distances(X1_d, sin_transform=True)
        #
        # print("Bond distance - mean:", mean_bond, "std:", std_bond, "max:", max_bond, "min:", min_bond)
        # print("Angle distance (sin) - mean:", mean_angle_sin, "std:", std_angle_sin, "max:", max_angle_sin, "min:", min_angle_sin)
        # print("Dihedral distance (sin) - mean:", mean_dihedral_sin, "std:", std_dihedral_sin, "max:", max_dihedral_sin, "min:", min_dihedral_sin)

        # Create empty kernel
        K_X1X2 = torch.empty(((1 + self.Nmask) * X1_b.shape[0],
                              (1 + self.Nmask) * X2_b.shape[0]), dtype=self.torch_data_type, device=self.device)

        # Expand value fingerprint
        X1_b_expanded = X1_b[:, None, :, None, :]  # [Ndata1, 1, Ncenter, 1, Nfeature]
        X2_b_expanded = X2_b[None, :, None, :, :]  # [1, Ndata2, 1, Ncenter, Nfeature]
        X1_b__outer_minus__X2_b = X1_b_expanded - X2_b_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]
        X2_b__outer_minus__X1_b = X2_b_expanded - X1_b_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]

        X1_a_expanded = X1_a[:, None, :, None, :]  # [Ndata1, 1, Ncenter, 1, Nfeature]
        X2_a_expanded = X2_a[None, :, None, :, :]  # [1, Ndata2, 1, Ncenter, Nfeature]
        X1_a__outer_minus__X2_a = X1_a_expanded - X2_a_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]
        X2_a__outer_minus__X1_a = X2_a_expanded - X1_a_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]

        X1_d_expanded = X1_d[:, None, :, None, :]  # [Ndata1, 1, Ncenter, 1, Nfeature]
        X2_d_expanded = X2_d[None, :, None, :, :]  # [1, Ndata2, 1, Ncenter, Nfeature]
        X1_d__outer_minus__X2_d = X1_d_expanded - X2_d_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]
        X2_d__outer_minus__X1_d = X2_d_expanded - X1_d_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]

        # [Ndata, Ndata, Ncenter, Ncenter]
        __k_X1X2_b = torch.exp(-torch.linalg.norm(X1_b__outer_minus__X2_b, dim=4) ** 2 / (2 * scale_b ** 2))
        __k_X2X1_b = torch.exp(-torch.linalg.norm(X2_b__outer_minus__X1_b, dim=4) ** 2 / (2 * scale_b ** 2))

        __k_X1X2_a = torch.exp(-torch.linalg.norm(torch.sin(X1_a__outer_minus__X2_a / 2), dim=4) ** 2 / (2 * scale_a ** 2))
        __k_X2X1_a = torch.exp(-torch.linalg.norm(torch.sin(X2_a__outer_minus__X1_a / 2), dim=4) ** 2 / (2 * scale_a ** 2))

        __k_X1X2_d = torch.exp(-torch.linalg.norm(torch.sin(X1_d__outer_minus__X2_d / 2), dim=4) ** 2 / (2 * scale_d ** 2))
        __k_X2X1_d = torch.exp(-torch.linalg.norm(torch.sin(X2_d__outer_minus__X1_d / 2), dim=4) ** 2 / (2 * scale_d ** 2))

        """"K_X1X2[:X1.shape[0], :X2.shape[0]] kernel between fp1 and fp2"""
        # [Ndata, Ndata, Ncenter, Ncenter] -> [Ndata, Ndata]
        K_X1X2[:X1_b.shape[0], :X2_b.shape[0]] = torch.sum(self.weight ** 2 * __k_X1X2_b * __k_X1X2_a * __k_X1X2_d,
                                                           dim=(2, 3))

        """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), :X2.shape[0]] kernel between fp1_deriv and fp2"""
        # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
        dX1_b_reshaped = dX1_b.flatten(start_dim=2, end_dim=3)
        dX1_a_reshaped = dX1_a.flatten(start_dim=2, end_dim=3)
        dX1_d_reshaped = dX1_d.flatten(start_dim=2, end_dim=3)

        intermediate_result_1 = torch.einsum('xycnf,xcbf->xycnb', X1_b__outer_minus__X2_b / scale_b ** 2, dX1_b_reshaped)
        intermediate_result_1 += 1 / 4 * torch.einsum('xycnf,xcbf->xycnb', torch.sin(X1_a__outer_minus__X2_a) / scale_a ** 2, dX1_a_reshaped)
        intermediate_result_1 += 1 / 4 * torch.einsum('xycnf,xcbf->xycnb', torch.sin(X1_d__outer_minus__X2_d) / scale_d ** 2, dX1_d_reshaped)

        K_X1X2[X1_b.shape[0]:, :X2_b.shape[0]].copy_(torch.einsum('xycn,xycnb->xyb', self.weight ** 2 * __k_X1X2_b * __k_X1X2_a * __k_X1X2_d,
                                                                  intermediate_result_1).permute(0, 2, 1).contiguous().view(X1_b.shape[0] * self.Nmask, X2_b.shape[0]))

        """K_X1X2[:X1.shape[0], X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] kernel between fp1 and fp2_deriv"""
        # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
        dX2_b_reshaped = dX2_b.flatten(start_dim=2, end_dim=3)
        dX2_a_reshaped = dX2_a.flatten(start_dim=2, end_dim=3)
        dX2_d_reshaped = dX2_d.flatten(start_dim=2, end_dim=3)

        intermediate_result_2 = torch.einsum('xycnf,yndf->xycnd', X2_b__outer_minus__X1_b / scale_b ** 2, dX2_b_reshaped)
        intermediate_result_2 += 1 / 4 * torch.einsum('xycnf,yndf->xycnd', torch.sin(X2_a__outer_minus__X1_a) / scale_a ** 2, dX2_a_reshaped)
        intermediate_result_2 += 1 / 4 * torch.einsum('xycnf,yndf->xycnd', torch.sin(X2_d__outer_minus__X1_d) / scale_d ** 2, dX2_d_reshaped)

        K_X1X2[:X1_b.shape[0], X2_b.shape[0]:].copy_(torch.einsum('xycn,xycnd->xyd', self.weight ** 2 * __k_X1X2_b * __k_X1X2_a * __k_X1X2_d,
                                                                  intermediate_result_2).view(X1_b.shape[0], X2_b.shape[0] * self.Nmask))

        """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] 
        kernel between fp1_deriv and fp2_deriv"""
        # size sign in eisum:
        # x, y: Ndata
        # c, n: Ncenter
        # b, d: Natom *3
        # f, f: Nfeature

        K_X1X2[X1_b.shape[0]:, X2_b.shape[0]:].copy_(torch.einsum('xycn,xycnbd->xybd', self.weight ** 2 * __k_X1X2_b * __k_X1X2_a * __k_X1X2_d,
                                                                  torch.einsum('xycnb,xycnd->xycnbd',
                                                                               intermediate_result_1,
                                                                               intermediate_result_2)).permute(0, 2, 1, 3).contiguous().view(X1_b.shape[0] * self.Nmask, X2_b.shape[0] * self.Nmask))

        intermediate_result_3 = torch.einsum('xcbf,yndf->xycnbd', dX1_b_reshaped / scale_b ** 2, dX2_b_reshaped)
        intermediate_result_3 += 1 / 4 * torch.einsum('xcbf,xycnf,yndf->xycnbd', dX1_a_reshaped / scale_a ** 2, torch.cos(X2_a__outer_minus__X1_a), dX2_a_reshaped)
        intermediate_result_3 += 1 / 4 * torch.einsum('xcbf,xycnf,yndf->xycnbd', dX1_d_reshaped / scale_d ** 2, torch.cos(X2_d__outer_minus__X1_d), dX2_d_reshaped)

        K_X1X2[X1_b.shape[0]:, X2_b.shape[0]:].add_(torch.einsum('xycn,xycnbd->xybd', self.weight ** 2 * __k_X1X2_b * __k_X1X2_a * __k_X1X2_d,
                                                                 intermediate_result_3).permute(0, 2, 1, 3).contiguous().view(X1_b.shape[0] * self.Nmask, X2_b.shape[0] * self.Nmask))

        return K_X1X2

    def kernel_matrix_batch(self, images, batch_size=25):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(X,X) i.e. full kernel matrix for training data.
        X = [Ntrain or Nsparse or Ntest, Ncenter, Nfeature]
        dX = [Ntrain or Nsparse or Ntest, Ncenter, Natom, 3, Nfeature]
        '''

        Ndata = len(images)
        X_N_batch = get_N_batch(Ndata, batch_size)
        X_indexes = get_batch_indexes_N_batch(Ndata, X_N_batch)

        K_XX = torch.empty((Ndata * (1 + self.Nmask),
                            Ndata * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, X_N_batch):
            fp_i, dfp_dr_i = self.generate_internal_descriptor(images=images[X_indexes[i][0]:X_indexes[i][1]])

            row_indexes = torch.arange(X_indexes[i][0], X_indexes[i][1])
            row_deriv_indexes = torch.arange(Ndata + X_indexes[i][0] * self.Nmask,
                                             Ndata + X_indexes[i][1] * self.Nmask)
            selected_rows = torch.cat([row_indexes, row_deriv_indexes])

            K_XX[selected_rows[:, None], selected_rows] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                 X2=fp_i, dX2=dfp_dr_i)

            for j in range(i + 1, X_N_batch):
                fp_j, dfp_dr_j = self.generate_internal_descriptor(images=images[X_indexes[j][0]:X_indexes[j][1]])

                col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
                col_deriv_indexes = torch.arange(Ndata + X_indexes[j][0] * self.Nmask,
                                                 Ndata + X_indexes[j][1] * self.Nmask)
                selected_cols = torch.cat([col_indexes, col_deriv_indexes])

                K_XX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                     X2=fp_j, dX2=dfp_dr_j)
                K_XX[selected_cols[:, None], selected_rows] = K_XX[selected_rows[:, None], selected_cols].T

        return K_XX

    def kernel_vector_batch(self, eval_images, train_images, batch_size=25):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(x,X) i.e. the kernel matrix between fingerprint "x" and the training data fingerprints in "X".
        x = [Ntest or Nsparse, Ncenter, Nfeature]
        dx = [Ntest or Nsparse, Ncenter, Natom, 3, Nfeature]
        X = [Ntrain, Ncenter, Nfeature]
        dX = [Ntrain, Ncenter, Natom, 3, Nfeature]
        '''

        Ntest = len(eval_images)
        Ntrain = len(train_images)

        x_N_batch = get_N_batch(Ntest, batch_size)
        x_indexes = get_batch_indexes_N_batch(Ntest, x_N_batch)

        X_N_batch = get_N_batch(Ntrain, batch_size)
        X_indexes = get_batch_indexes_N_batch(Ntrain, X_N_batch)

        K_xX = torch.empty((Ntest * (1 + self.Nmask),
                            Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, x_N_batch):
            fp_i, dfp_dr_i = self.generate_internal_descriptor(images=eval_images[x_indexes[i][0]:x_indexes[i][1]])

            row_indexes = torch.arange(x_indexes[i][0], x_indexes[i][1])
            row_deriv_indexes = torch.arange(Ntest + x_indexes[i][0] * self.Nmask,
                                             Ntest + x_indexes[i][1] * self.Nmask)
            selected_rows = torch.cat([row_indexes, row_deriv_indexes])

            for j in range(0, X_N_batch):
                fp_j, dfp_dr_j = self.generate_internal_descriptor(images=train_images[X_indexes[j][0]:X_indexes[j][1]])

                col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
                col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * self.Nmask,
                                                 Ntrain + X_indexes[j][1] * self.Nmask)
                selected_cols = torch.cat([col_indexes, col_deriv_indexes])

                K_xX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                     X2=fp_j, dX2=dfp_dr_j)

        return K_xX

    def set_params(self, params):
        '''
        Set new (hyper)parameters for the kernel function.
        '''
        self.update(params)
        self.kerneltype.update(params)



class InputParameters(object):
    """
    Information from the input file (train.in)
    train_file     :: [TRAININGSET] List of training set files
    test_file      :: [TESTPERCENT] List of test set files
    """

    def __init__(self):
        self.file_format       = 'xsf'
        self.device            = 'cpu'
        self.train_file        = []
        self.test_file         = []

        self.data_type = 'float64'
        self.data_process = 'batch'
        self.descriptor        = 'cartesian coordinates'
        self.standardization   = False
        self.mask_constraints  = False

        self.fit_weight        = True
        self.fit_scale         = True

        self.filter            = True
        self.filter_threshold  = 0.1

        self.soap_r_cut        = 5.0
        self.soap_n_max        = 6
        self.soap_l_max        = 4
        self.soap_sigma        = 0.5
        self.soap_rbf          = 'gto'
        self.soap_sparse       = False
        self.soap_centers      = None
        self.soap_method       = 'numerical'
        self.soap_n_jobs       = 1
        self.soap_param = {'r_cut': self.soap_r_cut,
                           'n_max': self.soap_n_max,
                           'l_max': self.soap_l_max,
                           'sigma': self.soap_sigma,
                           'rbf': self.soap_rbf,
                           'sparse': self.soap_sparse,
                           'centers': self.soap_centers,
                           'method': self.soap_method,
                           'n_jobs': self.soap_n_jobs}

        self.kerneltype        = 'sqexp'
        self.scale             = 0.4
        self.weight            = 1.0
        self.noise             = 1e-6
        self.noisefactor       = 0.5
        self.use_forces        = True
        self.sparse            = None
        self.sparse_derivative = None
        self.autograd          = False
        self.train_batch_size  = 25
        self.eval_batch_size   = 25
        self.get_variance      = True

        self.train_write       = False
        self.test_write        = False
        self.train_model_save  = False

        self.additional_write  = False
        self.disp_length       = 0.055
        self.num_copy          = 25

    def update_soap_param(self):
        self.soap_param = {'r_cut': self.soap_r_cut,
                           'n_max': self.soap_n_max,
                           'l_max': self.soap_l_max,
                           'sigma': self.soap_sigma,
                           'rbf': self.soap_rbf,
                           'sparse': self.soap_sparse,
                           'centers': self.soap_centers,
                           'method': self.soap_method,
                           'n_jobs': self.soap_n_jobs}



import torch


class ConstantPrior:
    '''Constant prior, with energy = constant and zero forces

    Parameters:

    constant: energy value for the constant.
    '''

    def __init__(self, constant, data_type=torch.float64, device='cpu', atoms_mask=None, **kwargs):
        self.data_type = data_type
        self.device = device
        self.constant = torch.tensor(constant, dtype=self.data_type, device=self.device)
        self.atoms_mask = atoms_mask

    def set_constant(self, constant):
        self.constant = constant

    def potential_per_data(self, image=None, get_forces=True):
        if get_forces:
            if self.atoms_mask is not None:
                Nmask = self.atoms_mask.shape[0]
            else:
                Nmask = len(image.get_chemical_symbols()) * 3
            output = torch.zeros((1 + Nmask,), dtype=self.data_type, device=self.device)

            output[0] = self.constant
            return output
        else:
            return self.constant

    def potential_batch(self, images=None, get_forces=True):
        if get_forces:
            if self.atoms_mask is not None:
                Nmask = self.atoms_mask.shape[0]
            else:
                Nmask = len(images[0].get_chemical_symbols()) * 3

            output_array = torch.zeros(((1 + Nmask) * len(images),), dtype=self.data_type, device=self.device)
            output_array[:len(images)] = self.constant
            return output_array
        else:
            output_array = torch.tensor([self.constant for i in range(len(images))], dtype=self.data_type, device=self.device)
            return output_array

    def update(self, images, Y, L, use_forces):
        """Update the constant to maximize the marginal likelihood.

        The optimization problem:
        m = argmax [-1/2 (y-m).T K^-1(y-m)]

        can be turned into an algebraic problem
        m = [ u.T K^-1 y]/[u.T K^-1 u]

        where u is the constant prior with energy 1 (eV).

        parameters:
        ------------
        X: training data
        Y: training targets
        L: Cholesky factor of the kernel """

        self.set_constant(torch.tensor(1.0, dtype=self.data_type, device=self.device))
        if use_forces:
            u = self.potential_batch(images)
        else:
            u = self.potential_batch(images, get_forces=False)

        # Solve the system L * L^T * v = Y (Cholesky solve)
        w = torch.cholesky_solve(u.view(-1, 1), L, upper=False)
        m = torch.matmul(w.view(-1), Y.flatten()) / torch.matmul(w.view(-1), u)

        self.set_constant(m)



import numpy as np
import random


def acquisition(train_images, candidates, mode='min_energy', objective='min'):
    """
    Acquisition function class.
    This function is in charge of ordering a list of Atoms candidates to
    be evaluated in order to minimize the number of function queries to get
    to the converged solution.

    Parameters
    ----------
    train_images: list
        List of Atoms containing the previously evaluated structures.
    candidates: list
        Unordered list of Atoms with the candidates to be evaluated.
    mode: string
        Name of the acquisition function in charge of ordering the candidates.
        Modes available are:
            - 'energy': Sort candidates by predicted energy.
            - 'uncertainty': Sort candidates by predicted uncertainty.
            - 'random': Randomly shuffles the candidates list.
            - 'ucb': Sort candidates by predicted energy plus uncertainty.
            - 'lcb': Sort candidates by predicted energy minus uncertainty.
    objective: string
         Objective of the acquisition function. Options:
            - 'min': If our goal is to minimize the target function.
            - 'max': If our goal is to maximize the target function.


    Returns
    -------
    Sorted candidates: list
        Ordered list of Atoms with the candidates to be evaluated.
    """

    x = []
    y = []
    pred_x = []
    pred_y = []
    pred_fmax = []
    pred_unc = []

    # Gather all required information to decide the order of the candidates.
    for i in candidates:
        pred_x.append(i.get_positions().reshape(-1))
        pred_y.append(i.get_potential_energy())
        if mode == 'uncertainty' or mode == 'ucb' or mode == 'lcb':
            pred_unc.append(i.calc.results['uncertainty'])
        if mode == 'fmax':
            pred_fmax.append(np.sqrt((i.get_forces()**2).sum(axis=1).max()))

    for i in train_images:
        x.append(i.get_positions().reshape(-1))
        y.append(i.get_potential_energy())

    implemented_acq = ['energy', 'fmax', 'uncertainty', 'ucb', 'lcb', 'random']

    if mode not in implemented_acq:
        msg = 'The selected acquisition function is not implemented. ' \
              'Implemented are: ' + str(implemented_acq)
        raise NotImplementedError(msg)

    if mode == 'energy':
        if objective == 'min':
            score_index = np.argsort(pred_y)
        if objective == 'max':
            score_index = list(reversed(np.argsort(pred_y)))

    if mode == 'uncertainty':
        if objective == 'min':
            score_index = np.argsort(pred_unc)
        if objective == 'max':
            score_index = list(reversed(np.argsort(pred_unc)))

    if mode == 'fmax':
        if objective == 'min':
            score_index = np.argsort(pred_fmax)
        if objective == 'max':
            score_index = list(reversed(np.argsort(pred_fmax)))

    if mode == 'ucb':
        e_plus_u = np.array(pred_y) + np.array(pred_unc)
        if objective == 'min':
            score_index = np.argsort(e_plus_u)
        if objective == 'max':
            score_index = list(reversed(np.argsort(e_plus_u)))

    if mode == 'lcb':
        e_minus_u = np.array(pred_y) - np.array(pred_unc)
        if objective == 'min':
            score_index = np.argsort(e_minus_u)
        if objective == 'max':
            score_index = list(reversed(np.argsort(e_minus_u)))

    if mode == 'random':
        ordered_index = list(range(len(candidates)))
        score_index = random.sample(ordered_index, len(ordered_index))

    # Order candidates (from best to worst candidates):
    ordered_images = []
    for i in score_index:
        ordered_images.append(candidates[i])

    return ordered_images



import sys
from setuptools import setup, find_packages

install_requires = ['dscribe', 'ase']

if __name__ == '__main__':

    assert sys.version_info >= (3, 0), 'python>=3 is required'

    with open('./README.md', 'rt', encoding='UTF8') as f:
        long_description = f.read()

    with open('aenet_gpr/__init__.py', 'r') as init_file:
        for line in init_file:
            if "__version__" in line:
                version = line.split()[2].strip('\"')
                break

setup(name='aenet-gpr',
      version=version,
      description='Atomistic simulation tools based on Gaussian Processes Regression',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='In Won Yeu',
      author_email='iy2185@columbia.edu',
      url='https://github.com/atomisticnet/aenet-gpr',
      license='MPL-2.0',
      packages=find_packages(),
      install_requires=install_requires,
      python_requires='>=3',
      keywords=['machine learning', 'potential energy surface', 'aenet', 'data augmentation'],
)



from .gpr_iterative import GaussianProcess
from .gpr_batch import GaussianProcess
from .gpr_batch_internal import GaussianProcessInternal
from .pytorch_kernel import FPKernel, FPKernelNoforces
from .pytorch_kernel_internal import FPKernelInternal
from .pytorch_kerneltypes import SquaredExp
from .calculator import GPRCalculator
from .prior import ConstantPrior

__all__ = ["gpr_iterative", "gpr_batch", "gpr_batch_internal", "pytorch_kernel", "pytorch_kernel_internal", "pytorch_kerneltypes", "calculator", "prior",
           "GaussianProcess", "GaussianProcessInternal", "FPKernel", "FPKernelNoforces", "FPKernelInternal", "SquaredExp", "GPRCalculator", "ConstantPrior"]



import os
import glob
import copy

import numpy as np
import torch

import ase.io
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from aenet_gpr.src import gpr_iterative
from aenet_gpr.src import gpr_batch
from aenet_gpr.util.prepare_data import standard_output, inverse_standard_output


class ReferenceData(object):
    def __init__(self, structure_files: list = None,
                 file_format: str = 'xsf',
                 device='cpu',
                 descriptor='cartesian coordinates',
                 standardization=False,
                 data_type='float64',
                 data_process='batch',
                 soap_param=None,
                 mask_constraints=False):

        self.data_process = data_process
        if data_type == 'float32':
            self.data_type = 'float32'
            self.numpy_data_type = np.float32
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.numpy_data_type = np.float64
            self.torch_data_type = torch.float64

        self.device = device

        self.images = []
        self.lattice = np.array([], dtype=self.numpy_data_type)  # [Ndata, 3, 3] if pbc
        self.structure = np.array([], dtype=self.numpy_data_type)  # [Ndata, Natom, 3]
        self.energy = np.array([], dtype=self.numpy_data_type)  # [Ndata]
        self.force = np.array([], dtype=self.numpy_data_type)  # [Ndata, Natom, 3]

        self.descriptor = descriptor
        self.standardization = standardization
        self.calculator = None

        self.fix_ind = None
        self.pbc = None
        self.species = None
        self.num_atom = None

        if structure_files is not None:
            self.read_structure_files(structure_files, file_format)
        self.set_data()

        self.energy_scale = np.array([], dtype=self.numpy_data_type)  # [Ndata]
        self.force_scale = np.array([], dtype=self.numpy_data_type)  # [Ndata, Natom, 3]

        if soap_param is None:
            self.soap_param = {'r_cut': 5.0,
                               'n_max': 6,
                               'l_max': 4,
                               'sigma': 0.5,
                               'rbf': 'gto',
                               'sparse': False,
                               'centers': None,
                               'method': 'numerical',
                               'n_jobs': 1}
        else:
            self.soap_param = soap_param

        self.mask_constraints = mask_constraints
        if self.mask_constraints:
            self.atoms_mask = self.create_mask()
        else:
            self.atoms_mask = None

    def read_structure_files(self, structure_files=None, file_format: str = 'xsf'):
        # check 1: force unit
        # check 2: self.images, self.energy, self.force consistency
        if file_format == 'xsf':
            structures = []
            energies = []
            forces = []
            for structure_file in structure_files:
                image, structure, energy, force = self.read_xsf_image(structure_file)
                self.images.extend(image)
                structures.append(structure)
                energies.append(energy)
                forces.append(force)

            self.structure = np.asarray(structures, dtype=self.numpy_data_type)

            if any(energy is None for energy in energies):
                self.energy = None
            else:
                self.energy = np.asarray(energies, dtype=self.numpy_data_type)

            if any(force is None for force in forces):
                self.force = None
            else:
                self.force = np.asarray(forces, dtype=self.numpy_data_type)

        elif file_format == 'ase':
            self.images = structure_files
            self.structure = np.asarray([image.get_positions() for image in self.images], dtype=self.numpy_data_type)
            self.energy = np.asarray([image.get_potential_energy() for image in self.images],
                                     dtype=self.numpy_data_type)
            self.force = np.asarray([image.get_forces() for image in self.images], dtype=self.numpy_data_type)

        else:
            for structure_file in structure_files:
                self.images.extend(ase.io.read(structure_file, index=':', format=file_format))

            self.structure = np.asarray([image.get_positions() for image in self.images], dtype=self.numpy_data_type)
            try:
                self.energy = np.asarray([image.get_potential_energy() for image in self.images],
                                         dtype=self.numpy_data_type)
            except:
                self.energy = None

            try:
                self.force = np.asarray([image.get_forces() for image in self.images], dtype=self.numpy_data_type)
            except:
                self.force = None

    def write_params(self):
        return dict(num_data=len(self.images),
                    calculator=self.calculator.hyper_params,
                    fix_ind=self.fix_ind,
                    pbc=self.pbc,
                    species=self.species,
                    num_atom=self.num_atom)

    def set_data(self):
        self.species = self.images[0].get_chemical_symbols()
        lattice_constants = []
        if np.all(self.images[0].get_pbc()):
            self.pbc = True
            lattice_constants.append(self.images[0].get_cell())

            for i, image in enumerate(self.images[1:]):
                species_ = image.get_chemical_symbols()
                assert species_ == self.species, "Chemical elements are not homogeneous with {0} data".format(i + 1)

                pbc_ = np.all(image.get_pbc())
                assert pbc_ == self.pbc, "Periodic boundary condition error: different pbc files are contained"

                lattice_constants.append(image.get_cell())

            self.lattice = np.asarray(lattice_constants, dtype=self.numpy_data_type)

        else:
            self.pbc = False

            for i, image in enumerate(self.images[1:]):
                species_ = image.get_chemical_symbols()
                assert species_ == self.species, "Chemical elements are not homogeneous with {0} data".format(i + 1)

                pbc_ = np.all(image.get_pbc())
                assert pbc_ == self.pbc, "Periodic boundary condition error: different pbc files are contained"

        self.num_atom = len(self.species)

    def create_mask(self):
        """
        This function mask atoms coordinates that will not participate in the
        model, i.e. the coordinates of the atoms that are kept fixed
        or constraint.
        """
        atoms = self.images[0]
        constraints = atoms.constraints
        mask_constraints = torch.ones_like(torch.tensor(atoms.positions), dtype=torch.bool)
        for i in range(0, len(constraints)):
            try:
                mask_constraints[constraints[i].a] = ~constraints[i].mask
            except Exception:
                pass

            try:
                mask_constraints[constraints[i].index] = False
            except Exception:
                pass

            try:
                mask_constraints[constraints[0].a] = ~constraints[0].mask
            except Exception:
                pass

            try:
                mask_constraints[constraints[-1].a] = ~constraints[-1].mask
            except Exception:
                pass
        return torch.argwhere(mask_constraints.reshape(-1)).reshape(-1)

    def set_fix_index_by_coord(self, max, axis=2):
        pos = self.images[0].get_positions()
        self.fix_ind = np.where(pos[:, axis] < max)

    def set_fix_index(self, fix_ind):
        self.fix_ind = fix_ind

    def generate_cartesian(self, images):

        fp = []
        for image in images:
            fp.append(torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(self.device))

        fp = torch.stack(fp).to(self.device)  # (Ndata, Natom*3)
        fp = fp.unsqueeze(1)  # (Ndata, 1, Natom*3)

        return fp

    def generate_cartesian_per_data(self, image):

        fp = torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(self.device)

        fp = fp.unsqueeze(0)  # (1, Natom*3)
        fp = fp.unsqueeze(0)  # (1, 1, Natom*3)

        return fp

    def filter_similar_data(self, threshold=0.1):
        """
        Remove training data that is too close below the threshold

        Parameters:
            threshold (float): minimum descriptor distance
        """
        X = self.generate_cartesian(self.images)
        N = X.shape[0]

        keep_indices = []
        remove_indices = []

        for i in range(N):
            xi = X[i].flatten()
            keep = True
            for idx in keep_indices:
                xj = X[idx].flatten()
                dist = torch.linalg.norm(xi - xj)
                if dist < threshold:
                    keep = False
                    break
            if keep:
                keep_indices.append(i)
            else:
                remove_indices.append(i)

        if remove_indices != []:
            print(f"Removed images {remove_indices} from training data (near-duplicates; risk of ill-conditioned covariance).")

            keep_images = [self.images[i] for i in keep_indices]
            self.read_structure_files(structure_files=keep_images, file_format='ase')
            self.set_data()

    def config_calculator(self, kerneltype='sqexp', scale=0.4, weight=1.0, noise=1e-6, noisefactor=0.5,
                          use_forces=True, sparse=None, sparse_derivative=None, autograd=False,
                          train_batch_size=25, eval_batch_size=25,
                          fit_weight=True, fit_scale=True):

        if self.standardization:
            Y_train_tensor = torch.as_tensor(self.energy_scale, dtype=self.torch_data_type).to(self.device)
            dY_train_tensor = torch.as_tensor(self.force_scale, dtype=self.torch_data_type).to(self.device)
        else:
            Y_train_tensor = torch.as_tensor(self.energy, dtype=self.torch_data_type).to(self.device)
            dY_train_tensor = torch.as_tensor(self.force, dtype=self.torch_data_type).to(self.device)

        if self.data_process == 'iterative':
            self.calculator = gpr_iterative.GaussianProcess(kerneltype=kerneltype,
                                                            scale=scale,
                                                            weight=weight,
                                                            noise=noise,
                                                            noisefactor=noisefactor,
                                                            use_forces=use_forces,
                                                            images=self.images,
                                                            function=Y_train_tensor,
                                                            derivative=dY_train_tensor,
                                                            sparse=sparse,
                                                            sparse_derivative=sparse_derivative,
                                                            autograd=autograd,
                                                            data_type=self.data_type,
                                                            device=self.device,
                                                            soap_param=self.soap_param,
                                                            descriptor=self.descriptor,
                                                            atoms_mask=self.atoms_mask).to(self.device)
        elif self.data_process == 'batch':
            self.calculator = gpr_batch.GaussianProcess(kerneltype=kerneltype,
                                                        scale=scale,
                                                        weight=weight,
                                                        noise=noise,
                                                        noisefactor=noisefactor,
                                                        use_forces=use_forces,
                                                        images=self.images,
                                                        function=Y_train_tensor,
                                                        derivative=dY_train_tensor,
                                                        sparse=sparse,
                                                        sparse_derivative=sparse_derivative,
                                                        autograd=autograd,
                                                        train_batch_size=train_batch_size,
                                                        eval_batch_size=eval_batch_size,
                                                        data_type=self.data_type,
                                                        device=self.device,
                                                        soap_param=self.soap_param,
                                                        descriptor=self.descriptor,
                                                        atoms_mask=self.atoms_mask).to(self.device)

        self.calculator.train_model()
        if fit_scale:
            self.fit_scale_only()
        if fit_weight:
            self.fit_weight_only(use_forces=True)

    def fit_weight_only(self, use_forces=True):
        """
        Fit weight of the kernel keeping all other hyperparameters fixed.
        Here we assume the kernel k(x,x',theta) can be factorized as:
        k = weight**2 * sqexp(x,x',other hyperparameters)
        """

        max_weight = 100.0
        prev_weight = copy.deepcopy(self.calculator.weight)

        try:
            if use_forces:
                _prior_array = self.calculator.prior.potential_batch(self.images, get_forces=True)
                _factor = torch.sqrt(torch.matmul(self.calculator.YdY.flatten() - _prior_array,
                                                  self.calculator.model_vector) / _prior_array.shape[0])
                self.calculator.weight *= _factor.squeeze()
                self.calculator.hyper_params.update(dict(weight=self.calculator.weight))
                self.calculator.kernel.set_params(self.calculator.hyper_params)

            else:
                _prior_array = self.calculator.prior.potential_batch(self.images, get_forces=False)
                _n_train = len(self.images)
                _factor = torch.sqrt(torch.matmul(self.calculator.YdY.flatten()[:_n_train] - _prior_array,
                                                  self.calculator.model_vector[:_n_train]) / _prior_array.shape[0])
                self.calculator.weight *= _factor.squeeze()
                self.calculator.hyper_params.update(dict(weight=self.calculator.weight))
                self.calculator.kernel.set_params(self.calculator.hyper_params)

            if self.calculator.weight < max_weight:
                pass
            else:
                raise ValueError(f"Weight parameter is too high ({self.calculator.weight}).")

        except Exception as e:
            print(f"{e}")
            print("Fix the weight parameter")
            self.calculator.weight = prev_weight
            self.calculator.hyper_params.update(dict(weight=self.calculator.weight))
            self.calculator.kernel.set_params(self.calculator.hyper_params)

        print('Updated weight:', self.calculator.weight.item())

        self.calculator.train_model()

        return

    def fit_scale_only(self, candidates=3, factor=2.0):
        """
        Update the kernel scale keeping all other hyperparameters fixed by evaluating marginal likelihood over candidate scales.

        Parameters:
        X: ndarray (training inputs)
        Y: ndarray (training targets)
        candidates: int (number of candidate scales)
        factor: float (multiplicative factor to generate candidate scales)

        Returns:
        Updated scale value
        """
        current_scale = self.calculator.scale
        best_scale = current_scale
        best_logp = -torch.inf

        # Generate candidate scales logarithmically spaced around current scale
        candidate_scales = current_scale * factor ** torch.linspace(-1, 1, candidates, device=self.device)

        print('Candidate scales:', candidate_scales)
        for candidate_scale in candidate_scales:
            # Set temporary scale
            self.calculator.scale = candidate_scale
            self.calculator.hyper_params.update(dict(scale=self.calculator.scale))
            self.calculator.kernel.set_params(self.calculator.hyper_params)

            # Train GP with current candidate scale
            try:
                self.calculator.train_model()
            except Exception as e:
                print(f"{e}")
                continue

            # Evaluate marginal likelihood
            y_flat = self.calculator.YdY.flatten()
            m_ = self.calculator.prior.potential_batch(self.images, get_forces=True)
            a_ = self.calculator.model_vector

            logP = -0.5 * torch.matmul(y_flat - m_, a_) - torch.sum(torch.log(torch.diag(self.calculator.K_XX_L))) \
                   - self.calculator.Ntrain * 0.5 * torch.log(torch.as_tensor(2 * torch.pi, dtype=self.torch_data_type).to(self.device))

            # MAP 보정: log-normal prior centered at previous (or initial) scale
            log_l = torch.log(candidate_scale)
            mu = torch.log(torch.tensor(0.4, dtype=self.torch_data_type, device=self.device))
            sigma = torch.tensor(0.5, dtype=self.torch_data_type, device=self.device)
            log_prior = -0.5 * ((log_l - mu) / sigma) ** 2 - log_l
            lambda_prior = 1.0  # 튜닝
            logP_total = logP + lambda_prior * log_prior

            # Keep track of best scale
            if logP_total > best_logp:
                best_logp = logP_total
                best_scale = candidate_scale

        # Update to best scale found
        self.calculator.scale = best_scale
        self.calculator.hyper_params.update(dict(scale=self.calculator.scale))
        self.calculator.kernel.set_params(self.calculator.hyper_params)
        print('Updated scale:', best_scale.item())

        # Train GP with best candidate scale
        self.calculator.train_model()

        return

    def evaluation(self, get_variance=False):

        self.calculator.eval()
        with torch.no_grad():
            if get_variance:
                energy_gpr_scale, force_gpr_scale, uncertainty_gpr = self.calculator(eval_images=self.images,
                                                                                     get_variance=get_variance)

                return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), uncertainty_gpr.cpu().detach().numpy()

            else:
                energy_gpr_scale, force_gpr_scale, _ = self.calculator(eval_images=self.images,
                                                                       get_variance=get_variance)

                return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), None

    def standardize_energy_force(self, reference_training_energy):
        """
        Y = [n_systems]
        dY = [n_systems, Natom, 3]
        :return:
        """
        self.energy_scale, self.force_scale = standard_output(reference_training_energy,
                                                              self.energy,
                                                              self.force)

    def inverse_standardize_energy_force(self, reference_training_energy):
        """
        Y = [n_systems]
        dY = [n_systems, Natom, 3]
        :return:
        """
        self.energy, self.force = inverse_standard_output(reference_training_energy,
                                                          self.energy_scale,
                                                          self.force_scale)

    def read_xsf_image(self, path):

        image = ase.io.read(path, index=':', format='xsf')
        with open(path, 'r') as infile:
            lines = infile.readlines()

        structure = np.empty((len(image[0]), 3), dtype=self.numpy_data_type)

        try:
            energy = np.asarray(lines[0].split()[4], dtype=self.numpy_data_type)
        except IndexError:
            energy = None
        except ValueError:
            energy = None

        force = np.empty((len(image[0]), 3), dtype=self.numpy_data_type)
        if "ATOMS" in lines[2]:
            for i, line in enumerate(lines[3:]):
                structure[i, :] = np.asarray(line.split()[1:4], dtype=self.numpy_data_type)
                try:
                    force[i, :] = np.asarray(line.split()[4:], dtype=self.numpy_data_type)
                except:
                    force = None

        elif "CRYSTAL" in lines[2]:
            for i, line in enumerate(lines[9:]):
                structure[i, :] = np.asarray(line.split()[1:4], dtype=self.numpy_data_type)
                try:
                    force[i, :] = np.asarray(line.split()[4:], dtype=self.numpy_data_type)
                except:
                    force = None

        image[0].calc = SinglePointCalculator(image[0].copy(), energy=energy, forces=force)

        return image, structure, energy, force

    def write_image_xsf(self, path):

        i = 0
        for image in self.images:
            ase.io.write(os.path.join(path, "file_{0:0>5}.xsf".format(i)), image, format="xsf")
            i = i + 1

        files = glob.glob(os.path.join(path, "file_*.xsf"))
        files.sort()

        for i, file in enumerate(files):
            with open(file, 'r') as infile:
                lines = infile.readlines()

            if self.pbc:
                new_lines = lines[:7]
                del lines[:7]
            else:
                new_lines = [lines[0]]
                del lines[0]

            for j, line in enumerate(lines):
                tmp = line.split()
                tmp[0] = self.species[j]

                try:
                    tmp[4] = "%16.14f" % self.force[i, j, 0]
                    tmp[5] = "%16.14f" % self.force[i, j, 1]
                    tmp[6] = "%16.14f" % self.force[i, j, 2]
                except:
                    tmp.append("%16.14f" % self.force[i, j, 0])
                    tmp.append("%16.14f" % self.force[i, j, 1])
                    tmp.append("%16.14f" % self.force[i, j, 2])

                new_line = "     ".join(tmp)
                new_lines.append(new_line + "\n")

            with open(file, "w") as outfile:
                comment = "# total energy = %20.16f eV\n\n" % float(self.energy[i])
                outfile.write("%s" % comment)

                for new_line in new_lines:
                    outfile.write(new_line)

    def save_data(self, file="data_dict.pt"):
        """
        self.data_type
        self.numpy_data_type
        self.torch_data_type

        self.images -> self.structures (numpy)
        self.energy (numpy)
        self.force (numpy)

        self.device
        self.descriptor
        self.standardization
        self.calculator -> calculator.save

        self.fix_ind
        self.pbc
        self.species
        self.num_atom

        self.soap -> self.soap.parameters
        """

        state = {
            'data_type': self.data_type,
            'data_process': self.data_process,
            'numpy_data_type': self.numpy_data_type,
            'torch_data_type': self.torch_data_type,
            'lattice': torch.as_tensor(self.lattice),
            'structure': torch.as_tensor(self.structure),
            'energy': torch.as_tensor(self.energy),
            'force': torch.as_tensor(self.force),
            'device': self.device,
            'descriptor': self.descriptor,
            'standardization': self.standardization,
            'fix_ind': self.fix_ind,
            'pbc': self.pbc,
            'species': self.species,
            'num_atom': self.num_atom,
            'soap_param': self.soap_param,
        }
        torch.save(state, file)
        self.calculator.save_data()

    def load_data(self, file="data_dict.pt"):
        state = torch.load(file)

        self.data_process = state.get('data_process')
        self.data_type = state.get('data_type')
        self.numpy_data_type = state.get('numpy_data_type')
        self.torch_data_type = state.get('torch_data_type')

        self.images = []
        self.lattice = state.get('lattice').cpu().detach().numpy()
        self.structure = state.get('structure').cpu().detach().numpy()
        self.energy = state.get('energy').cpu().detach().numpy()
        self.force = state.get('force').cpu().detach().numpy()

        self.device = state.get('device')
        self.descriptor = state.get('descriptor')
        self.standardization = state.get('standardization')

        self.fix_ind = state.get('fix_ind')
        self.pbc = state.get('pbc')
        self.species = state.get('species')
        self.num_atom = state.get('num_atom')

        self.soap_param = state.get('soap_param')

        if self.pbc:
            atom = Atoms(self.species)
            atom.set_pbc(self.pbc)
            for i, pos in enumerate(self.structure):
                atom.set_positions(pos)
                atom.set_cell(self.lattice[i])
                self.images.append(copy.deepcopy(atom))

        else:
            atom = Atoms(self.species)
            atom.set_pbc(self.pbc)
            for i, pos in enumerate(self.structure):
                atom.set_positions(pos)
                self.images.append(copy.deepcopy(atom))

        if self.data_process == 'iterative':
            self.calculator = gpr_iterative.GaussianProcess(images=self.images,
                                                            data_type=self.data_type,
                                                            device=self.device,
                                                            soap_param=self.soap_param,
                                                            descriptor=self.descriptor).to(self.device)
            self.calculator.load_data()

        elif self.data_process == 'batch':
            self.calculator = gpr_batch.GaussianProcess(images=self.images,
                                                        data_type=self.data_type,
                                                        device=self.device,
                                                        soap_param=self.soap_param,
                                                        descriptor=self.descriptor).to(self.device)
            self.calculator.load_data()



import numpy as np
import torch
from ase.calculators.calculator import Calculator, all_changes


class GPRCalculator(Calculator):

    implemented_properties = ['energy', 'forces']

    def __init__(self, calculator, train_data, **kwargs):
        Calculator.__init__(self, **kwargs)

        self.calculator = calculator
        self.train_data = train_data
        self.results = {}

    def calculate(self, atoms=None,
                  properties=['energy', 'forces', 'uncertainty'],
                  system_changes=all_changes):
        '''
        Calculate the energy and forces for a given Atoms structure.
        Predicted energies can be obtained by *atoms.get_potential_energy()*,
        predicted forces using *atoms.get_forces()*
        '''

        Calculator.calculate(self, atoms, properties, system_changes)

        pred, kernel = self.calculator.eval_data_per_data(eval_image=atoms)
        energy_gpr = pred[0].cpu().detach().numpy()
        if self.train_data.mask_constraints:
            force_gpr = torch.zeros((len(atoms) * 3), dtype=pred.dtype)
            force_gpr[self.train_data.atoms_mask] = pred[1:]
            force_gpr = force_gpr.view(len(atoms), 3).cpu().detach().numpy()
        else:
            force_gpr = pred[1:].view(len(atoms), 3).cpu().detach().numpy()

        var = self.calculator.eval_variance_per_data(get_variance=True, eval_image=atoms, k=kernel)
        uncertainty_gpr = torch.sqrt(var[0, 0]) / self.calculator.weight
        uncertainty_gpr = uncertainty_gpr.cpu().detach().numpy()

        if self.train_data.standardization:
            mean_energy = np.mean(self.train_data.energy)
            std_energy = np.std(self.train_data.energy)

            # Restore Energy: scaled_energy_target * std + mean
            energy_gpr = energy_gpr * std_energy + mean_energy

            # Restore Force: scaled_force_target * std
            force_gpr = force_gpr * std_energy

        self.results['energy'] = energy_gpr
        self.results['forces'] = force_gpr
        self.results['uncertainty'] = uncertainty_gpr



import datetime
import time


def io_print(text):
    print(text, flush=True)


def io_current_time():
    aux = str(datetime.datetime.now())[:-6]
    io_print(aux)


def io_data_read_finalize(t, mem_CPU, mem_GPU, flag, energy_shape, force_shape):

    io_print("")
    io_print("Read reference {0} data".format(flag))
    io_print("")
    io_print("Time needed for reading data:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    print("Energy data size:", energy_shape, " # (N_data, )")
    print("Force data size:", force_shape, " # (N_data, N_atom, 3)")
    io_line()


def io_set_descriptor_finalize(t, mem_CPU, mem_GPU, flag, descriptor, fp_shape, dfp_dr_shape):

    io_line()
    io_print("Generate {0} descriptors".format(flag))
    io_print("")
    io_print("Time needed for setting descriptor:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    print("Descriptor:", descriptor)
    print("Descriptor size:", fp_shape, " # (N_data, N_center, N_feature)")
    print("Descriptor derivatives size:", dfp_dr_shape, " # (N_data, N_center, N_atom, 3, N_feature)")
    io_line()


def io_train_finalize(t, mem_CPU, mem_GPU, data_param: dict):

    io_line()
    io_print("Model train")
    io_train_parameters(data_param=data_param)
    io_print("Time needed for training:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_line()


def io_train_parameters(data_param: dict):

    io_print("")
    io_print("Training parameters")
    io_print(data_param)
    io_print("")


def io_train_model_save(t, mem_CPU, mem_GPU):

    io_line()
    io_print("Trained model save")
    io_print("")
    io_print("Saving training data to [data_dict.pt]")
    io_print("Saving GPR model parameters to [calc_dict.pt]")
    io_print("")
    io_print("Time needed for saving:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_double_line()
    io_print("")
    io_print("")


def io_train_model_load(t, mem_CPU, mem_GPU):

    io_double_line()
    io_print("Model load")
    io_print("")
    io_print("Loading Training data from [data_dict.pt]")
    io_print("Loading GPR model parameters from [calc_dict.pt]")
    io_print("")
    io_print("Time needed for loading:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_double_line()


def io_test_evaluation(t, mem_CPU, mem_GPU, data_param: dict):

    io_line()
    io_print("Model evaluation for test set")
    io_test_parameters(data_param=data_param)
    io_print("Time needed for test evaluation:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_line()


def io_test_parameters(data_param: dict):
    io_print("")
    io_print("Test parameters")
    io_print(data_param)
    io_print("")


def io_additional_generate_finalize(t, mem_CPU, mem_GPU, disp_length, num_copy, num_reference, num_additional):

    io_line()
    io_print("Additional structure generation")
    io_print("")
    io_print("Time needed for additional generation:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    print("Displacement length (Ang):", disp_length)
    print("Multiple:", num_copy)
    print("N_additional = N_reference ({0}) * Multiple ({1}) = {2}".format(num_reference, num_copy, num_additional))
    io_line()


def io_additional_evaluation(t, mem_CPU, mem_GPU, data_param: dict):
    io_line()
    io_print("Model evaluation for additional set")
    io_test_parameters(data_param=data_param)
    io_print("Time needed for additional evaluation:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_line()


def io_additional_parameters(data_param: dict):
    io_print("")
    io_print("Augmentation parameters")
    io_print(data_param)
    io_print("")


def io_train_write_finalize(t, mem_CPU, mem_GPU, path):

    io_double_line()
    io_print("Writing train xsf files to {0}".format(path))
    io_print("")
    io_print("Time needed for writing xsf files:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_double_line()


def io_test_write_finalize(t, mem_CPU, mem_GPU, path):

    io_double_line()
    io_print("Writing test xsf files to {0}".format(path))
    io_print("")
    io_print("Time needed for writing xsf files:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_double_line()


def io_additional_write_finalize(t, mem_CPU, mem_GPU, path):

    io_double_line()
    io_print("Writing additional xsf files to {0}".format(path))
    io_print("")
    io_print("Time needed for writing xsf files:  {: 18.6f} s".format(time.time() - t))
    io_print("Maximum CPU memory used:   {: 18.6f} GB".format(mem_CPU))
    io_print("Maximum GPU memory used:   {: 18.6f} GB".format(mem_GPU))
    io_double_line()


def io_line():
    io_print("----------------------------------------------------------------------")


def io_print_title(text):
    io_double_line()
    io_print(text)
    io_double_line()


def io_double_line():
    io_print("======================================================================")


def io_print_header():
    io_double_line()
    io_print("aenet-GPR: surrogate GPR for GPR-ANN indirect force training")
    io_double_line()
    io_print("")
    io_current_time()
    io_print("")
    io_print("Developed by In Won Yeu")
    io_print("")
    io_print("This program performs three main steps:")
    io_print("1. Train: Generates a GPR model using the provided structure, energy, and force data.")
    io_print("2. Test: Uses the generated GPR model to predict values for the test set structures.")
    io_print("3. Augmentation: Performs data augmentation in xsf file format, compatible with aenet-(PyTorch),")
    io_print("supporting a GPR-ANN training in conjunction with aenet-(PyTorch).")
    io_print("")
    io_print("Each of these steps is executed once the input file (train.in) contains the keywords:")
    io_print("Train_file [train file path]")
    io_print("Test_file [test file path]")
    io_print("Additional_write [True]")
    # io_print("Once the Train step is completed, the generated GPR model is saved in [data_dict.pt] and [calc_dict.pt].")
    # io_print("Using these saved model files, you can later run only the Test or Augmentation steps separately.")
    # io_print("This program is distributed in the hope that it will be useful,")
    # io_print("but WITHOUT ANY WARRANTY; without even the implied warranty of")
    # io_print("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the")
    # io_print("GNU General Public License in file 'LICENSE' for more details.")
    io_print("")
    io_print("")

